@@ -5,11 +5,12 @@
 
 __all__ = ['fixture', 'fixture_dirname', 'TestCase']
 
+import json
+import logging
 import os.path
 import unittest
-import json
-
-import logging
+from functools import wraps
+from contextlib import contextmanager
 
 logging.basicConfig(level=logging.INFO)
 
@@ -23,5 +24,13 @@ def fixture(filename):
         return json.load(file)
 
 
+@contextmanager
+def move_cwd():
+    cwd = os.getcwd()
+    os.chdir(here)
+    yield cwd
+    os.chdir(cwd)
+
+
 class TestCase(unittest.TestCase):
     pass
